
from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class Reactor(AppConfig):
    name = 'reactor'
    verbose_name = 'Django Reactor'

    def ready(self):
        from . import auto_broadcast  # noqa
        autodiscover_modules('live')



import simdjson
from typing import Generator

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


parser = simdjson.Parser()


def loads(text_data):
    return parser.parse(text_data)


def dumps(obj):
    return simdjson.dumps(obj, cls=Encoder)


class Encoder(DjangoJSONEncoder):

    def default(self, o):
        if isinstance(o, models.Model):
            return o.pk

        if isinstance(o, models.QuerySet):
            return list(o.values_list('pk', flat=True))

        if isinstance(o, (Generator, set)):
            return list(o)

        if hasattr(o, '__json__'):
            return o.__json__()

        try:
            import numpy as n
        except ImportError:
            pass
        else:
            number = (
                n.ndarray,
                n.float,
                n.float16,
                n.float32,
                n.float64,
                n.float128
            )
            if isinstance(o, number):
                return o.tolist()

        return super().default(o)




from django.dispatch import receiver
from django.db import models
from django.db.models.signals import post_save, pre_delete, m2m_changed

from .utils import broadcast
from .settings import AUTO_BROADCAST


MODEL = AUTO_BROADCAST.get('MODEL', False)
MODEL_PK = AUTO_BROADCAST.get('MODEL_PK', False)
RELATED = AUTO_BROADCAST.get('RELATED', False)
M2M = AUTO_BROADCAST.get('M2M', False)


if MODEL or MODEL_PK or RELATED:
    @receiver(post_save)
    def broadcast_post_save(sender, instance, created=False, **kwargs):
        name = sender._meta.model_name
        if MODEL:
            broadcast(name)
            created and broadcast(f'{name}.new')

        if instance.pk is not None:
            MODEL_PK and broadcast(f'{name}.{instance.pk}')
            RELATED and broadcast_related(sender, instance, created=created)

    @receiver(pre_delete)
    def broadcast_post_delete(sender, instance, **kwargs):
        name = sender._meta.model_name
        MODEL and broadcast(name)

        if instance.pk is not None:
            MODEL and broadcast(f'{name}.del')
            MODEL_PK and broadcast(f'{name}.{instance.pk}')
            RELATED and broadcast_related(sender, instance, deleted=True)


def broadcast_related(sender, instance, deleted=False, created=False):
    for field in sender._meta.get_fields():
        if isinstance(field, models.ForeignKey):
            fk_id = getattr(instance, field.attname)
            if fk_id is not None:
                fk_model_name = field.related_model._meta.model_name
                fk_attr_name = field.related_query_name()
                group_name = f'{fk_model_name}.{fk_id}.{fk_attr_name}'
                broadcast(group_name)
                if created:
                    broadcast(f'{group_name}.new')
                if deleted:
                    broadcast(f'{group_name}.del')
        elif M2M and isinstance(field, models.ManyToManyField):
            fk_attr_name = field.related_query_name()
            if fk_attr_name != '+':
                fk_ids = getattr(instance, field.attname).values_list(
                    'id', flat=True
                )
                fk_model_name = field.related_model._meta.model_name
                group_names = [
                    f'{fk_model_name}.{fk_id}.{fk_attr_name}'
                    for fk_id in fk_ids
                ]
                broadcast(*group_names)
                if created:
                    broadcast(*[f'{gn}.new' for gn in group_names])
                if deleted:
                    broadcast(*[f'{gn}.del' for gn in group_names])


if M2M:
    @receiver(m2m_changed)
    def broadcast_m2m_changed(
        sender, instance, action, model, pk_set, **kwargs
    ):
        if action.startswith('post_') and instance.pk:
            model_name = model._meta.model_name
            attr_name = get_name_of(sender, model)
            updates = [f'{model_name}.{pk}.{attr_name}' for pk in pk_set or []]
            broadcast(*updates)
            broadcast(*[f'{u}.{instance.pk}' for u in updates])

            model = type(instance)
            model_name = model._meta.model_name
            attr_name = get_name_of(sender, model)
            update = f'{model_name}.{instance.pk}.{attr_name}'
            broadcast(update)
            broadcast(*[f'{update}.{pk}' for pk in pk_set or []])


def get_name_of(through, model):
    for model_field in model._meta.get_fields():
        found = (
            getattr(model_field, 'through', None) or
            getattr(getattr(model, model_field.name, None), 'through', None)
        )
        if through is found:
            return model_field.name



import difflib
from uuid import uuid4
from functools import reduce

from django.shortcuts import resolve_url
from django.template.loader import get_template, select_template
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.functional import cached_property

from . import settings
from . import json
from .utils import send_to_channel


class RootComponent(dict):

    def __init__(self, context):
        super().__init__()
        self._context = context

    @cached_property
    def _channel_name(self):
        return self._context.get('channel_name')

    def get_or_create(self, _name, _parent_id=None, id=None, **state):
        id = str(id or '')
        component: Component = self.get(id)
        if component:
            component._refresh(**state)
        else:
            component = Component._build(
                _name,
                _context=self._context,
                _root_component=self,
                _parent_id=_parent_id,
                id=id,
            )
            component.mount(**state)
            self[component.id] = component
        return component

    def pop(self, id, default=None):
        return super().pop(id, default)

    def dispatch_user_event(self, name, state):
        component: Component = self.get(state['id'])
        if component:
            return component._dispatch(name, state)
        else:
            send_to_channel(self._channel_name, 'remove', id=state['id'])

    def propagate_update(self, event):
        origin = event['origin']
        for component in list(self.values()):
            if origin in component._subscriptions:
                component._update(**event)
                html_diff = component._render_diff()
                yield {'id': component.id, 'html_diff': html_diff}


class Component:
    template_name = ''
    template = None
    extends = 'div'
    _all = {}

    def __init_subclass__(cls, name=None, public=True):
        if public:
            name = name or cls.__name__
            name = ''.join([('-' + c if c.isupper() else c) for c in name])
            name = name.strip('-').lower()
            cls._all[name] = cls
            cls._tag_name = name
        return super().__init_subclass__()

    def __str__(self):
        state = escape(json.dumps(self.serialize()))
        return mark_safe(
            f'is="{self._tag_name}" '
            f'id="{self.id}" '
            f'state="{state}"'
        )

    def header(self):
        return str(self)

    # Constructors

    @classmethod
    def _build(cls, tag_name, *args, **kwargs):
        return cls._all[tag_name](*args, **kwargs)

    def __init__(
        self,
        _context,
        _root_component: RootComponent = None,
        _parent_id=None,
        id=None,
    ):
        self._context = _context
        self._destroy_sent = False
        self._is_frozen = False
        self._redirected_to = None
        self._last_sent_html = []
        if _root_component is None:
            self._root_component = RootComponent(_context)
        else:
            self._root_component = _root_component
        self._parent_id = _parent_id
        self._subscriptions = set()
        self.id = str(id or uuid4())

    # User events

    def subscribe(self, *room_names):
        for room_name in room_names:
            if room_name not in self._subscriptions:
                self._subscriptions.add(room_name)
                send_to_channel(
                    self._channel_name,
                    'subscribe',
                    room_name=room_name
                )

    def unsubscribe(self, room_name):
        self._subscriptions.discard(room_name)

    @cached_property
    def _channel_name(self):
        return self._context.get('channel_name')

    def _dispatch(self, name, args=None):
        getattr(self, f'receive_{name}')(**(args or {}))
        return self._render_diff()

    # State persistence & front-end communication

    def freeze(self, *args, **kwargs):
        self._is_frozen = True

    def _update(self, **kwargs):
        """Entrypoint for broadcast events"""
        return self._refresh()

    def _refresh(self, **state):
        self.mount(**dict(self.serialize(), **state))

    def mount(self, **state):
        """
        Override so given an initial state be able to  re-create or update the
        component state
        """
        pass

    def serialize(self):
        """
        Override to send this state to the front-end and calling `mount` for
        state recreation
        """
        return {'id': self.id}

    # Redirects & rendering

    def send_destroy(self):
        self._destroy_sent = True
        send_to_channel(self._channel_name, 'remove', id=self.id)

    def send_redirect(self, url,  *args, **kwargs):
        url = resolve_url(url, *args, **kwargs)
        if self._channel_name:
            send_to_channel(self._channel_name, 'push_state', url=url)
            self.freeze()
        else:
            self._redirected_to = url

    def send_parent(self, _name, **kwargs):
        if self._parent_id:
            self.send(_name, id=self._parent_id, **kwargs)

    def send(self, _name, id=None, **kwargs):
        send_to_channel(
            self._channel_name,
            'send_component',
            name=_name,
            state=dict(kwargs, id=id or self.id),
        )

    def _render_diff(self):
        html = self._render()
        html = html.splitlines()
        if html and self._last_sent_html != html:
            if settings.USE_HTML_DIFF:
                diff = []
                for x in difflib.ndiff(self._last_sent_html, html):
                    indicator = x[0]
                    if indicator == ' ':
                        diff.append(1)
                    elif indicator == '+':
                        diff.append(x[2:])
                    elif indicator == '-':
                        diff.append(-1)

                if diff:
                    diff = reduce(compress_diff, diff[1:], diff[:1])
            else:
                diff = html
            self._last_sent_html = html
            return diff

    def _render(self):
        if self._is_frozen:
            html = self._last_sent_html
        elif self._destroy_sent:
            html = ''
        elif self._redirected_to:
            html = (
                f'<meta'
                f' http-equiv="refresh"'
                f' content="0; url={self._redirected_to}"'
                f'>'
            )
        else:
            html = self._get_template().render(self._get_context()).strip()
        return mark_safe(html)

    def _get_template(self):
        if not self.template:
            if isinstance(self.template_name, (list, tuple)):
                self.template = select_template(self.template_name)
            else:
                self.template = get_template(self.template_name)
        return self.template

    def _get_context(self):
        return dict(
            {
                attribute: getattr(self, attribute)
                for attribute in dir(self)
                if not attribute.startswith('_')
            },
            this=self,
        )


class AuthComponent(Component, public=False):

    def mount(self, *args, **kwargs):
        if self.user.is_authenticated:
            # Listen to user logout and refresh
            return True
        else:
            self.send_destroy()

    @cached_property
    def user(self):
        return self._context['user']


class StaffComponent(AuthComponent, public=False):
    def mount(self, *args, **kwargs):
        if super().mount() and self.user.is_staff:
            return True
        else:
            self.send_destroy()


def compress_diff(diff, diff_item):
    if isinstance(diff_item, str) or isinstance(diff[-1], str):
        diff.append(diff_item)
    else:
        same_sign = not (diff[-1] > 0) ^ (diff_item > 0)
        if same_sign:
            diff[-1] += diff_item
        else:
            diff.append(diff_item)
    return diff



import logging
from functools import wraps

from asgiref.sync import async_to_sync

from channels.layers import get_channel_layer

log = logging.getLogger('reactor')


def broadcast(*names, **kwargs):
    for name in names:
        log.debug(f'<-> {name}')
        send_to_group(name, 'update', **kwargs)


def on_commit(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        from django.db.transaction import on_commit
        on_commit(lambda: f(*args, **kwargs))
    return wrapper


def send_to_channel(_channel_name, type, **kwargs):
    if _channel_name:
        @on_commit
        def send_message():
            async_to_sync(get_channel_layer().send)(
                _channel_name, dict(type=type, **kwargs)
            )
        send_message()


def send_to_group(_whom, type, **kwargs):
    if _whom:
        @on_commit
        def send_message():
            async_to_sync(get_channel_layer().group_send)(
                _whom, dict(type=type, origin=_whom, **kwargs)
            )
        send_message()



import json
from django import template
from django.template.loader import render_to_string


from ..component import Component
from ..json import Encoder
from ..settings import INCLUDE_TURBOLINKS

register = template.Library()


@register.simple_tag
def reactor_header():
    return render_to_string(
        'reactor_header.html',
        {'include_tlinks': INCLUDE_TURBOLINKS}
    )


@register.simple_tag(takes_context=True)
def component(context, _name, id=None, **kwargs):
    parent = context.get('this')  # type: Component
    if parent:
        component = parent._root_component.get_or_create(
            _name,
            _parent_id=parent.id,
            id=id,
            **kwargs
        )
    else:
        component = Component._build(_name, _context=context, id=id)
        component.mount(**kwargs)
    return component._render()


@register.filter
def concat(value, arg):
    return f'{value}-{arg}'


@register.filter()
def tojson(value, indent=None):
    return json.dumps(value, cls=Encoder, indent=indent)


@register.filter
def eq(value, other):
    if value == other:
        return 'yes'
    else:
        return ''


@register.filter
def then(value, true_result):
    if value:
        return true_result
    else:
        return ''


@register.filter
def ifnot(value, false_result):
    if not value:
        return false_result
    else:
        return ''
